@@ -23,8 +23,8 @@ ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.
 */
 
-#if (defined(__linux__) || defined(__NetBSD__) || defined(__FreeBSD__)) && \
-	defined(__i386__)
+#if defined(__i386__) && (defined(__linux__) || defined(__NetBSD__) \
+	|| defined(__FreeBSD__) || defined(__OpenBSD__))
 
 #include <stdio.h>
 #include <string.h>
@@ -32,12 +32,13 @@ OTHER DEALINGS IN THE SOFTWARE.
 #if defined(__linux__)
 
 #include <asm/vm86.h>
+#include <signal.h>
 
 #ifdef USE_LIBC_VM86
 #include <sys/vm86.h>
 #endif
 
-#elif defined(__NetBSD__) || defined(__FreeBSD__)
+#elif defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
 
 #include <sys/param.h>
 #include <signal.h>
@@ -46,7 +47,11 @@ OTHER DEALINGS IN THE SOFTWARE.
 #include <machine/vm86.h>
 #include <machine/sysarch.h>
 
-#endif /* __NetBSD__ || __FreeBSD__ */
+#endif /* __NetBSD__ || __FreeBSD__ || __OpenBSD__ */
+
+#if defined(__FreeBSD__)
+#include <sys/ucontext.h>
+#endif
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -72,31 +77,63 @@ static struct {
 } mem_info = { 0 };
 
 static int
-real_mem_init(void)
+read_file(char *name, void *p, size_t n)
+{
+	int fd;
+
+	fd = open(name, O_RDONLY);
+
+	if (fd == -1) {
+		perror("open");
+		return 0;
+	}
+
+	if (read(fd, p, n) != n) {
+		perror("read");
+		close(fd);
+		return 0;
+	}
+
+	close(fd);
+
+	return 1;
+}
+
+static int
+map_file(void *start, size_t length, int prot, int flags, char *name, long offset)
 {
 	void *m;
-	int fd_zero;
+	int fd;
 
+	fd = open(name, (flags & MAP_SHARED) ? O_RDWR : O_RDONLY);
+
+	if (fd == -1) {
+		perror("open");
+		return 0;
+	}
+
+	m = mmap(start, length, prot, flags, fd, offset);
+
+	if (m == (void *)-1) {
+		perror("mmap");
+		close(fd);
+		return 0;
+	}
+
+	close(fd);
+	return 1;
+}
+
+static int
+real_mem_init(void)
+{
 	if (mem_info.ready)
 		return 1;
 
-	fd_zero = open("/dev/zero", O_RDONLY);
-	if (fd_zero == -1) {
-		perror("open /dev/zero");
-		return 0;
-	}
-
-	m = mmap((void *)REAL_MEM_BASE, REAL_MEM_SIZE,
+	if (!map_file((void *)REAL_MEM_BASE, REAL_MEM_SIZE,
 	 PROT_READ | PROT_WRITE | PROT_EXEC,
-	 MAP_FIXED | MAP_PRIVATE, fd_zero, 0);
-
-	if (m == (void *)-1) {
-		perror("mmap /dev/zero");
-		close(fd_zero);
+	 MAP_FIXED | MAP_PRIVATE, "/dev/zero", 0))
 		return 0;
-	}
-
-	close(fd_zero);
 
 	mem_info.ready = 1;
 	mem_info.count = 1;
@@ -203,7 +240,7 @@ LRMI_free_real(void *m)
 
 #if defined(__linux__)
 #define DEFAULT_VM86_FLAGS 	(IF_MASK | IOPL_MASK)
-#elif defined(__NetBSD__) || defined(__FreeBSD__)
+#elif defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
 #define DEFAULT_VM86_FLAGS  (PSL_I | PSL_IOPL)
 #define TF_MASK         PSL_T
 #define VIF_MASK        PSL_VIF
@@ -214,7 +251,7 @@ LRMI_free_real(void *m)
 #if defined(__linux__)
 #define CONTEXT_REGS	context.vm.regs
 #define REG(x)			x
-#elif defined(__NetBSD__)
+#elif defined(__NetBSD__) || defined(__OpenBSD__)
 #define CONTEXT_REGS	context.vm.substr.regs
 #define REG(x)			vmsc.sc_ ## x
 #elif defined(__FreeBSD__)
@@ -226,7 +263,7 @@ static struct {
 	int ready;
 	unsigned short ret_seg, ret_off;
 	unsigned short stack_seg, stack_off;
-#if defined(__linux__) || defined(__NetBSD__)
+#if defined(__linux__) || defined(__NetBSD__) || defined(__OpenBSD__)
 	struct vm86_struct vm;
 #elif defined(__FreeBSD__)
 	struct {
@@ -234,7 +271,7 @@ static struct {
 		ucontext_t uc;
 	} vm;
 #endif
-#if defined(__NetBSD__) || defined(__FreeBSD__)
+#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
 	int success;
 	jmp_buf env;
 	void *old_sighandler;
@@ -279,7 +316,6 @@ int
 LRMI_init(void)
 {
 	void *m;
-	int fd_mem;
 
 	if (context.ready)
 		return 1;
@@ -291,39 +327,26 @@ LRMI_init(void)
 	 Map the Interrupt Vectors (0x0 - 0x400) + BIOS data (0x400 - 0x502)
 	 and the ROM (0xa0000 - 0x100000)
 	*/
-	fd_mem = open("/dev/mem", O_RDWR);
-
-	if (fd_mem == -1) {
-		real_mem_deinit();
-		perror("open /dev/mem");
-		return 0;
-	}
-
-	m = mmap((void *)0, 0x502,
+	if (!map_file((void *)0, 0x502,
 	 PROT_READ | PROT_WRITE | PROT_EXEC,
-	 MAP_FIXED | MAP_PRIVATE, fd_mem, 0);
-
-	if (m == (void *)-1) {
-		close(fd_mem);
+	 MAP_FIXED | MAP_PRIVATE, "/dev/zero", 0)) {
 		real_mem_deinit();
-		perror("mmap /dev/mem");
 		return 0;
 	}
 
-	m = mmap((void *)0xa0000, 0x100000 - 0xa0000,
-	 PROT_READ | PROT_WRITE,
-	 MAP_FIXED | MAP_SHARED, fd_mem, 0xa0000);
-
-	if (m == (void *)-1) {
+	if (!read_file("/dev/mem", (void *)0, 0x502)) {
 		munmap((void *)0, 0x502);
-		close(fd_mem);
 		real_mem_deinit();
-		perror("mmap /dev/mem");
 		return 0;
 	}
 
-	close(fd_mem);
-
+	if (!map_file((void *)0xa0000, 0x100000 - 0xa0000,
+	 PROT_READ | PROT_WRITE,
+	 MAP_FIXED | MAP_SHARED, "/dev/mem", 0xa0000)) {
+		munmap((void *)0, 0x502);
+		real_mem_deinit();
+		return 0;
+	}
 
 	/*
 	 Allocate a stack
@@ -352,7 +375,7 @@ LRMI_init(void)
 #if defined(__linux__)
 	memset(&context.vm.int_revectored, 0, sizeof(context.vm.int_revectored));
 	set_bit(RETURN_TO_32_INT, &context.vm.int_revectored);
-#elif defined(__NetBSD__)
+#elif defined(__NetBSD__) || defined(__OpenBSD__)
 	set_bit(RETURN_TO_32_INT, &context.vm.int_byuser);
 #elif defined(__FreeBSD__)
 	set_bit(RETURN_TO_32_INT, &context.vm.init.int_map);
@@ -773,9 +796,19 @@ static int
 run_vm86(void)
 {
 	unsigned int vret;
+	sigset_t all_sigs, old_sigs;
+	unsigned long old_gs, old_fs;
 
 	while (1) {
+		// FIXME: may apply this to BSD equivalents?
+		sigfillset(&all_sigs);
+		sigprocmask(SIG_SETMASK, &all_sigs, &old_sigs);
+		asm volatile ("mov %%gs, %0" : "=rm" (old_gs));
+		asm volatile ("mov %%fs, %0" : "=rm" (old_fs));
 		vret = lrmi_vm86(&context.vm);
+		asm volatile ("mov %0, %%gs" :: "rm" (old_gs));
+		asm volatile ("mov %0, %%fs" :: "rm" (old_fs));
+		sigprocmask(SIG_SETMASK, &old_sigs, NULL);
 
 		if (VM86_TYPE(vret) == VM86_INTx) {
 			unsigned int v = VM86_ARG(vret);
@@ -805,8 +838,8 @@ run_vm86(void)
 
 	return 0;
 }
-#elif defined(__NetBSD__) || defined(__FreeBSD__)
-#if defined(__NetBSD__)
+#elif defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
+#if defined(__NetBSD__) || defined(__OpenBSD__)
 static void
 vm86_callback(int sig, int code, struct sigcontext *sc)
 {
@@ -906,7 +939,7 @@ run_vm86(void)
 		return (0);
 	}
 
-#if defined(__NetBSD__)
+#if defined(__NetBSD__) || defined(__OpenBSD__)
 	context.old_sighandler = signal(SIGURG, (void (*)(int))vm86_callback);
 #elif defined(__FreeBSD__)
 	context.old_sighandler = signal(SIGBUS, (void (*)(int))vm86_callback);
@@ -921,7 +954,7 @@ run_vm86(void)
 	}
 
 	if (setjmp(context.env)) {
-#if defined(__NetBSD__)
+#if defined(__NetBSD__) || defined(__OpenBSD__)
 		(void) signal(SIGURG, context.old_sighandler);
 #elif defined(__FreeBSD__)
 		(void) signal(SIGBUS, context.old_sighandler);
@@ -934,7 +967,7 @@ run_vm86(void)
 		return (0);
 	}
 
-#if defined(__NetBSD__)
+#if defined(__NetBSD__) || defined(__OpenBSD__)
 	if (i386_vm86(&context.vm) == -1)
 		return (0);
 #elif defined(__FreeBSD__)
@@ -948,7 +981,7 @@ run_vm86(void)
 	/* NOTREACHED */
 	return (0);
 }
-#endif	/* __NetBSD__ || __FreeBSD__ */
+#endif	/* __NetBSD__ || __FreeBSD__ || __OpenBSD__ */
 
 int
 LRMI_call(struct LRMI_regs *r)
@@ -1027,4 +1060,6 @@ LRMI_int(int i, struct LRMI_regs *r)
 	return vret;
 }
 
-#endif /* (__linux__ || __NetBSD__ || __FreeBSD__) && __i386__ */
+#else /* (__linux__ || __NetBSD__ || __FreeBSD__ || __OpenBSD__) && __i386__ */
+#warning "LRMI is not supported on your system!"
+#endif
